// +---------------------------------------------------------------------------+
//
//      Program:    humidity_senosr.c
//
//      Purpose:    Reading the humidity and temperature from an AM2302 sensor
//      
//      Target:     ATtiny13A
//
//      Author:     Herbert Dingfelder
//
//      License:    GPL 3
//
//      History:    12.04.2015  Initial Coding
//                              Reading from the AM2302 sensor works, checksum is correct
//                  13.04.2015  Added I2C routines in preparation for adding a LCD
//                  15.04.2015  Added supply voltage measuring via ADC
//                              Added power save mode by clocking down the system clock
//                  10.05.2015  Added even better power save with sleep mode watchdog
//                                triggered wakeup
//                              Added supply voltage dependend contrast compensation for 
//                                LCD by using different pulse width to negative voltage
//                                generation circuit
//                              Tested ok, works fine. => Version 1.0 released
//
// +---------------------------------------------------------------------------+

#define F_CPU 9600000
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/cpufunc.h>


// Compiler switches for debugging and adjustments
//#define ADC_HEX_READ
//#define FAST_VALUE_UPDATES

#define SENSOR_PORT PORTB
#define SENSOR_DDR  DDRB
#define SENSOR_IN   PINB
#define SENSOR_BIT  _BV(4)

#define ADC_CHANNEL 3               // ADC3 is connected to PB3          

#define pull_data_line_low          SENSOR_DDR |= SENSOR_BIT
#define release_data_line_to_high   SENSOR_DDR &=~SENSOR_BIT

#define wait_till_data_line_high    while((SENSOR_IN & SENSOR_BIT)==0)
#define wait_till_data_line_low     while((SENSOR_IN & SENSOR_BIT)!=0)

#define led_on                      LED_PORT |=  LED_BIT
#define led_off                     LED_PORT &= ~LED_BIT

#define i2c_bus                     PORTB
#define i2c_dir                     DDRB

#define i2c_data                    _BV(2)          // I2C data is PB2
#define i2c_clock                   _BV(1)          // I2C clock is PB1
#define i2c_data_and_clock          (i2c_data+i2c_clock) 

#define I2C_DELAY                   10              // number of microseconds between events on I2C bus

#define LCD_CONTRAST_PORT           PORTB
#define LCD_CONTRAST_DDR            DDRB
#define LCD_CONTRAST_BIT            _BV(0)          // LCD contrast generation on PB0

char videoram[36];                  // 36 bytes the lcd videoram as global variable

void init_lcd(void);
void update_lcd(void);
void i2c_start(void);
void i2c_stop(void);
void i2c_send(unsigned char sendbyte);
unsigned char asc2lcd(unsigned char asc_char);
void reset_handler(void);



// ============================= main program starts here ======================
int main(void)
{
    
    unsigned char bit_number;
    unsigned char byte_number;
    unsigned char sensor_byte;
    unsigned char sensor_information[5];
    unsigned char checksum;
    unsigned char checksum_ok=0;
    unsigned int relative_humidity_times10;
    unsigned int temperature_times10;
    unsigned char temperature_negative=0;
    unsigned int dummy;
    unsigned char digit;
    unsigned char hex2asc(unsigned char hexbyte);
    unsigned char contrast;
    unsigned char g,t;
    
   
    reset_handler();                // do all the regular port and register settings
    
    
    for(;;)
    {
                
        pull_data_line_low;         // send start signal to sensor by pulling the data line
        _delay_ms(1);               // to low for 1ms, then release it back to high
        release_data_line_to_high;
        
        wait_till_data_line_high;   // start of 20-40µs period called 'host pulls up'
        wait_till_data_line_low;    // end of 20-40µs period called 'host pulls up'
        wait_till_data_line_high;   // end of 80µs period called 'sensor pulls low'
        wait_till_data_line_low;    // end of 80µs period called 'sensor pulls up' = start of transmission
        
        checksum = 0;               // clear checksum for next calculation
        checksum_ok=0;              // reset the checksum ok flag
        
        for( byte_number = 0 ; byte_number < 5 ; byte_number++ )    // read 5 bytes from sensor
        {
        
            sensor_byte = 0;
            
            for( bit_number = 0 ; bit_number <= 7 ; bit_number++)   // read 8 bits per byte
            {
                
                wait_till_data_line_low; 
                wait_till_data_line_high;       // wait for a raising edge (beginning of bit)
                
                _delay_us(50);                  // 0 bits are 26.28µs long according to data sheet of sensor,
                                                // 1 bits are 70µs long, so we sample at 50µs
                
                sensor_byte <<= 1;              // shift already received byte to the left                           
                
                if( SENSOR_IN & SENSOR_BIT )    // if bit is set
                {                
                    sensor_byte |= 0x01;        // set the LSB                    
                }
                
            }
            
            sensor_information[byte_number] = sensor_byte;                  // store the read byte
        
        }
        
        
        for( byte_number = 0 ; byte_number < 4 ; byte_number++ )            // for the first 4 bytes
        {
            checksum = 0xFF & (checksum+sensor_information[byte_number]);   // add up last 8 bits
        }
        
        
        if( checksum == sensor_information[4] )         // compare calculated with received checksum
        {    
            checksum_ok = 1;
        }
        else
        {    
            checksum_ok = 0;
        }
            
        
        if( checksum_ok )                       // if checksum was ok, calculate numeric value
        {                                       //   for the temperature and the humidity
            
            relative_humidity_times10 = (sensor_information[0]<<8) + sensor_information[1];
            
            temperature_times10 =       (sensor_information[2]<<8) + sensor_information[3];

            if(temperature_times10 & 0x8000)    // if MSB in 16bit word set => temperature is negative
            {
                temperature_negative = 1;
            }
            else
            {
                temperature_negative = 0;
            }
            
            temperature_times10 &=~0x8000;      // mask out the MSB of the 16bit word 
         
        }
        else                                    // if checksum was not ok, set all value to 0xff
        {
            relative_humidity_times10 = 0xffff;         
            temperature_times10 = 0xffff;
            temperature_negative = 0xff;
        }
              
       
        videoram[0] = 'T';                      // start to fill the videoram with text and results
        videoram[1] = 'e';
        videoram[2] = 'm';
        videoram[3] = 'p';
        videoram[4] = ':';
        videoram[5] = ' ';
        
        dummy = temperature_times10;            // split the number into decimal digits...
        
        if(dummy>=1000)                         // limit the number to 3 digits
            dummy=999;
        
        digit = (unsigned char)(dummy/100);
        if(digit)                               // write digit to LCD only if it is not zero
        {
            videoram[6] = digit + 48;           
            if(temperature_negative)            // in case of sub-freezing temperatures
                videoram[5] = '-';
        }
        else
        {
            videoram[6] = ' ';                  // suppress leading zero
            if(temperature_negative)            // in case of sub-freezing temperatures
                videoram[6] = '-';       
        }
        
        dummy = dummy % 100;
        digit = (unsigned char)(dummy/10);
        videoram[7] = digit + 48;
        
        videoram[8] = ',';
        
        digit = dummy % 10;
        videoram[9] = digit +48;
        
        videoram[10] = 'C';
        videoram[11] = ' ';
        
        
        videoram[12] = 'r';                     // the second LCD text line starts
        videoram[13] = 'H';
        videoram[14] = 'u';
        videoram[15] = 'm';
        videoram[16] = ':';
        videoram[17] = ' ';
        
        dummy = relative_humidity_times10;
        
        digit = (unsigned char)(dummy/100);
        if(digit)
            videoram[18] = digit + 48;
        else
            videoram[18] = ' ';                  // suppress leading zero
        
        dummy = dummy % 100;
        digit = (unsigned char)(dummy/10);
        videoram[19] = digit + 48;
        
        videoram[20] = ',';
        
        digit = dummy % 10;
        videoram[21] = digit +48;
        
        videoram[22] = '%';
        videoram[23] = ' ';   
        
        
        // --- measure the supply voltage using the internal ADC ---
        
        // the voltage divider that feeds a fraction of the supply voltage to the
        // ADC input is part of the I2C pull-up resistor on the I2C clock line,
        // therefore we need to pull the I2C clock pin low, to have the voltage
        // divider grounded on the lower end (upper end is at Vcc anyway)
        i2c_dir |=  i2c_clock;              // clock-dir bit sets clock to output->low at pin
        
        ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // ADC prescaler :128, that 
                                            // gives ADC frequency of 9.6MHz/128 = 75kHz
        
        ADMUX = _BV(REFS0) | ADC_CHANNEL;   // select internal band gap as reference
                                            // and chose the ADC channel
        
        ADCSRA |= _BV(ADEN);                // switch on the ADC in general
        
        ADCSRA |= _BV(ADSC);                // start a single ADC conversion
        while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

        // data sheet recommends to discard first conversion, so we do one more
        ADCSRA |= _BV(ADSC);                // start a single ADC conversion
        while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete
        
        dummy = ADC;                        // take over ADC reading result into variable
        
        ADCSRA = 0;                         // completely disable the ADC to save power
            
        i2c_dir &= ~i2c_clock;              // release I2C clock line again for normal I2C work

        // from the derived supply voltage range, we set the optimal LCD contrast value
        
        contrast = 8;                       // for mid range voltages, this value is a good compromise
        
        if( dummy > 0x385 )                 // above 5.0V we use the smallest value, that is ok till 5.5V
            contrast = 5;

        if( dummy < 0x2F0 )                 // below 4.4V we use a rather high value, this ok down to 3.8V
            contrast = 15;
        
        if( dummy < 0x27B )                 // below 3.8V we use an even higher value, that works 
            contrast = 80;                  // down to 3.3V (which is the sensible lower limit,
                                            // as the humidity sensor does not work any lower)

#ifndef ADC_HEX_READ        
        
        dummy = dummy *48 + 7300;           // approximation formula, empirically derived, to calculate
                                            // the supply voltage from the read ADC value
        
        videoram[24] = 'V';
        videoram[25] = 'b';
        videoram[26] = 'a';
        videoram[27] = 't';
        videoram[28] = ':';
        videoram[29] = ' ';
        
        digit = (unsigned char)(dummy/10000);// split the supply voltage value into decimal digits
        videoram[30] = digit + 48;
        
        videoram[31] = ',';
        
        dummy = dummy % 10000;
        digit = (unsigned char)(dummy/1000);
        videoram[32] = digit + 48;
        
        dummy = dummy % 1000;
        digit = (unsigned char)(dummy/100);
        videoram[33] = digit + 48;
        
        videoram[34] = 'V';   
        videoram[35] = ' ';   

#endif

// the following is code that is only used for debugging and adjustments, it is not compiled for the productive code
// (shows the ADC readout value directly in DAC steps, not converted to volts)
#ifdef ADC_HEX_READ
        videoram[24] = hex2asc( contrast>>4 & 0x000F );
        videoram[25] = hex2asc( contrast    & 0x000F );
        videoram[26] = ' ';
        videoram[27] = 'A';
        videoram[28] = 'D';
        videoram[29] = 'C';
        videoram[30] = '=';
        videoram[31] = '0';
        videoram[32] = 'x';
        videoram[33] = hex2asc( dummy>>8 & 0x000F );
        videoram[34] = hex2asc( dummy>>4 & 0x000F );
        videoram[35] = hex2asc( dummy    & 0x000F );

#endif
         
        update_lcd();                   // trigger to send all videoram data to the LCD
        
        CLKPR = 0x80;                   // set clock prescaler change enable 
        CLKPR = 0x08;                   // set prescaler to clock division factor 256
                                        // => system clock runs at 37.5kHz
                                        // this reduces the current consumption a lot
 
        // for about 30 seconds, the processor mainly sleeps to save power, with very short
        // wakeups every 250ms, for about 12ms, to send a pulse into the contrast voltage generation
        // circuit for the LCD dispaly
        
#ifdef FAST_VALUE_UPDATES
        for(t=0; t< 4; t++ )            // for debugging and HW adjustments, readout every second
#else
        for(t=0; t< 115; t++ )          // normal usage, readout every 115 x 262ms = 30 seconds
#endif        
        {
                    
            LCD_CONTRAST_PORT |= LCD_CONTRAST_BIT;  // take the output pin to high
            
            // put the processor into sleep for 250ms, settings were preconfigured in reset_handler()
            
            MCUCR = (1<<SE) | (1<<SM1); // prepare for sleep mode by setting sleep enable and sleep mode 1 
                                        // (power down)
            wdt_reset();                // restart watchdog, to get the full watchdog timer time as sleep periode
            sleep_cpu();                // put the cpu to sleep
   
            
            // put out a pulse to the circuit that generates the negative voltage for the LCD contrast
            LCD_CONTRAST_PORT &= ~LCD_CONTRAST_BIT;
            
            for(g=0; g<contrast; g++)   // the longer the pulse, the darker the LCD contrast
                _delay_us(5);           // (due to downclocked system, one 'µs' is really 256µs long)
            
        }
        
        CLKPR = 0x80;                   // set clock prescaler change enable 
        CLKPR = 0x00;                   // set prescaler to clock division factor 1
                                        // => system clock runs at full 9.6MHz
 
    }
    
    return 0;
    
}



// --------------------------------- reset handler -----------------------------
void reset_handler(void)
{
    
     
    CLKPR = 0x80;                   // set clock prescaler change enable 
    CLKPR = 0x00;                   // set prescaler to clock division factor 1
                                    // => system clock runs at the full 9.6MHz
    
   // ----- from here on we preconfigure the watchdog for causing an interupt after 256ms -----
   // Since the interupt service routine is empty, a normal program execution will not be noticable affected.
   // This configuration can be used to put the cpu to sleep and have it woken up after 256ms.
   // A sensible sleep call looks like this:
   // MCUCR = (1<<SE) | (1<<SM1);       // prepare for sleep mode by setting sleep enable and sleep mode 1 (power down)
   // wdt_reset();                      // restart watchdog, to get the full watchdog timer time as sleep periode
   // sleep_cpu();                      // put the cpu to sleep
   
    cli();                                                  // Turn off global interrupt

    wdt_reset();                                            // Reset Watchdog Timer

    WDTCR |= (1<<WDCE) | (1<<WDE);                          // Start timed sequence (mandatory to 
                                                            // enable changes of the watchdog settings)

    // ---- Got four cycles to set the new values for the watchdog -------------
    WDTCR = (1<<WDTIE) | (1<<WDP2);                         // interupt mode with a timeout setting 250ms

    sei();                                                  // Turn on global interrupt

    // ----- watchdog is preconfigured now, sleep mode can be started with sleep_cpu() -----
                                    
    _delay_ms(3000);                // 3sec wait time at power on, to make sure that sensor is accesible
                                    // and to help new flasing of AVR, since the other signals
                                    // that are shared with the ISP interfact tend to confuse
                                    // the programmer during normal operation
                                       
    
    
    
    SENSOR_PORT &= ~SENSOR_BIT;     // switch output to zero, it will always stay zero,
                                    // the line is an open collector line with external
                                    // pull up, to pull it low, we switch the DDR bit
                                    // (data direction register) between input (high impedance)
                                    // to out (low impedance, hence pulling low)
                                    
    SENSOR_DDR &= ~SENSOR_BIT;      // for a start we stay high impedance, i.e not pulling low
  
    LCD_CONTRAST_DDR |= LCD_CONTRAST_BIT;   // switch contrast gneration pin to output  
    
    init_lcd();                     // initialize the LCD display via the I2C bus
    
}


//---------------------- initialize the LCD display --------------
void init_lcd(void)
{

    #define adr   0x74  // I2C address of the display
    #define cntr  0x80  // Ctrl-Byte -> Command follows
    #define cmd1  0x0C  // Command: Display on
    #define cmd2  0xA0  // set cursor to 1st position in display
    #define cntr2 0x40  // Ctrl-Byte -> normal characters for displaying will follow
    
    i2c_start();        // send an start-command on i2c-bus
            
    i2c_send(adr);      // address the display on i2c bus
    
    i2c_send(cntr);     // Ctrl-Byte -> Command follows
    
    i2c_send(cmd1);     // Command: Display on
    
    i2c_send(cntr);     // Ctrl-Byte -> Command follows
    
    i2c_send(cmd2);     // set cursor to 1st position in display
    
    i2c_send(cntr2);    // normal data follows
        
}

//------------------------ i2c start command ---------------------
void i2c_start(void)
{
        
    //set the databits itself to low, outputs will we controlled
    //by switching the data direction register
    //(output->low at pin, input->extern 10k resistor will create a high at pin)
    i2c_bus &= ~i2c_data_and_clock;      
        
    _delay_us(I2C_DELAY);                           // i2c-slower 
        
    // data and clock to high
    i2c_dir &= ~i2c_data_and_clock;                 // clearing data+clock dir. bit sets them to input->high at pin)
    
    _delay_us(I2C_DELAY);                           // i2c-slower 
        
    // take data low first 
    i2c_dir |=  i2c_data;                           // setting data-dir bit sets data to output->low at pin)
                        
    _delay_us(I2C_DELAY);                           // i2c-slower 
                        
    // then take clock low               
    i2c_dir |=  i2c_clock;                          // setting clock-dir bit sets clock to output->low at pin)
    
    _delay_us(I2C_DELAY);                           // i2c-slower 
    
}

//------------------------ i2c stop command ----------------------
void i2c_stop(void)
{

    // first take clock high again               
    i2c_dir &= ~i2c_clock;                          //clearing clock-dir bit sets clock to input->high at pin)
     
    _delay_us(I2C_DELAY);                           // i2c-slower 
    
    // then take data high again 
    i2c_dir &= ~i2c_data;                           //clearing data-dir bit sets data to input->high at pin)
    
    _delay_us(I2C_DELAY);                           // i2c-slower 
    
}

//--------------------- send byte via i2c bus ---------------------
void i2c_send(unsigned char sendbyte)
{
    unsigned char dummy;

    for(dummy=0; dummy<=7; dummy++)                 // send all 8 bits 
    {
    
        if( sendbyte & 0x80 )                       // if msb of sendbyte is high ->
        {
             i2c_dir &= ~i2c_data;                  // clearing data-dir. releases the data 
                                                    // pin and it snaps to high
        }
        else                                        // otherwise
        {
            i2c_dir |=  i2c_data;                   // pull data back to low
        }
                                                
        _delay_us(I2C_DELAY);                       // i2c-slower 
                                        
        i2c_dir &= ~i2c_clock;                      // shortly release the clock pin for a clock pulse
        _delay_us(I2C_DELAY);                       // i2c-slower 

        i2c_dir |=  i2c_clock;          
        _delay_us(I2C_DELAY);                       // i2c-slower 
                
        sendbyte <<= 1;                             // logic shift left to shift the next
                                                    // bit to the msb position
        
    }           
    
    i2c_dir &= ~i2c_data;                           // clearing data-dir. releases the data 
                                                    // pin and it snaps to high 

    // send one additional pulse on the clock-line for i2c-acknowledge 
    i2c_dir &= ~i2c_clock;                          // shortly release the clock pin for a clock pulse
    _delay_us(I2C_DELAY);                           // i2c-slower 
    i2c_dir |=  i2c_clock;      
    _delay_us(I2C_DELAY);                           // i2c-slower 
        
}

//--------------------------- ascii to lcd ------------------------------
// converts a ascii character into the lcd-character
// works for A..Z, a..z, 0..9, space and spec. char /()!%=
unsigned char asc2lcd(unsigned char asc_char)
{
    
    asc_char += 128;                                // all normal characters have ASC+128 as LCD code 
         
    return asc_char;

}        

// transfers the content of the videoram to the LCD
void update_lcd(void)
{
    unsigned char char_cnt;
    
    init_lcd();                                     // switch on lcd, clear screen and set cursor home

    
    for(char_cnt=0; char_cnt < 36; char_cnt++)      // write to all 3 lines, all 12 coloums
        i2c_send( asc2lcd(videoram[char_cnt]) );    // write the char. to LCD display
                                                    //   before convert ASCII -> LCD
    i2c_stop();

}

#ifdef ADC_HEX_READ
//--------------------------- hex to asc ---------------------------------
// converts a hex value (00..0F) into the ASCII-character '0'..'F'
unsigned char hex2asc(unsigned char hexbyte)
{
    
    hexbyte += 48;                                  // ASCII code for 0-9 is 48-57
    
    if(hexbyte>=58)                                 // ASCII code for A-F is 65-70 -> in that case add 7
        hexbyte += 7;
    
    return hexbyte;
}
#endif

// --------------------- Watchdog Timeout Interrupt --------------------
ISR( WDT_vect )             
{
    // user code here... (we don't need any, interupt is for wake-up of sleep only)
    
    _NOP();                                         // NOP = no operation, to avoid that compiler
                                                    // takes this away completely
    
}

